"""IPSec Crypto Configurations"""

from prismasase import return_auth
from prismasase.configs import Auth
from prismasase.restapi import prisma_request

def ipsec_crypto_profiles_get(ipsec_crypto_profile: str, folder: dict, **kwargs) -> str:
    """Checks if IPSec Crypto Profile Exists

    Args:
        ipsec_crypto_profile (str): _description_

    Returns:
        bool: _description_
    """
    auth: Auth = return_auth(**kwargs)
    ipsec_crypto_profile_id: str = ""
    # ipsec_crypto_profile_exists: bool = False
    params = folder
    ipsec_crypto_profiles = prisma_request(token=auth,
                                           url_type='ipsec-crypto-profiles',
                                           method="GET",
                                           params=params,
                                           verify=auth.verify)
    for entry in ipsec_crypto_profiles['data']:
        if entry['name'] == ipsec_crypto_profile:
            # ipsec_crypto_profile_exists = True
            ipsec_crypto_profile_id = entry['id']
    return ipsec_crypto_profile_id



"""Prisma Access SASE"""
from os.path import expanduser, exists

import yaml

from prismasase.configs import Config, Auth
from prismasase.utilities import set_bool

config = Config()
#if not all([config.CLIENT_ID, config.CLIENT_SECRET, config.TSG]):
home = expanduser("~")
filename = f"{home}/.config/.prismasase"
if exists(filename):
    with open(filename, 'r', encoding='utf-8') as yam:
        yaml_config = yaml.safe_load(yam)
    config.CLIENT_ID = yaml_config['CLIENT_ID']
    config.CLIENT_SECRET = yaml_config['CLIENT_SECRET']
    config.TSG = yaml_config['TSG']
    config.CERT = yaml_config.get('CERT', False)
else:
    config.CLIENT_ID = ""
    config.CLIENT_SECRET = ""
    config.TSG = ""
    config.CERT = "false"
    #config.CLIENT_ID = input("Please input Client ID: ")
    #config.CLIENT_SECRET = getpass("Please input Client Secret: ")
    #config.TSG = input("Please enter TSG ID: ")
    #config.CERT = input("Please enter custom cert location" +
    #                    "('true'|'false'|<custom_cert_location>): ")
config.CERT = set_bool(config.CERT)  # type: ignore

def return_auth(**kwargs) -> Auth:
    """_summary_

    Returns:
        Auth: _description_
    """
    auth = kwargs.pop('auth') if kwargs.get('auth') else ""
    if not auth:
        # print(f"DEBUG: {config.to_dict()}")
        auth = Auth(config.TSG, config.CLIENT_ID, config.CLIENT_SECRET, verify=config.CERT)
    return auth



"""Tags"""

import json

from prismasase import return_auth
from prismasase.configs import Auth
from prismasase.exceptions import (SASEError, SASEObjectExists)
from prismasase.restapi import prisma_request
from prismasase.statics import FOLDER, TAG_COLORS
from prismasase.utilities import default_params


def tags_list(folder: str, **kwargs) -> dict:
    """List out all tags in the specified folder

    Args:
        folder (str): _description_

    Returns:
        dict: _description_
    """
    auth: Auth = return_auth(**kwargs)
    params = default_params(**kwargs)
    params = {**FOLDER[folder], **params}
    data = []
    count = 0
    response = {
        'data': [],
        'offset': 0,
        'total': 0,
        'limit': 0
    }
    # Loops through to get all data in specified folder depending on limit and totals
    while (len(data) < response['total']) or count == 0:
        response = prisma_request(token=auth,
                                  method="GET",
                                  url_type='tags',
                                  params=params,
                                  verify=auth.verify)
        data = data + response['data']
        # Adjust to get the next set until you get all values
        params = {**params, **{'offset': params['offset'] + params['limit']}}
        count += 1
        # print(f"DEBUG: {params=}|{response=}|{count=}|{data=}|{len(data)}")
    response['data'] = data
    return response


def tags_create(folder: str, tag_name: str, **kwargs) -> dict:
    """Create Tag

    Args:
        folder (str): _description_

    Returns:
        dict: _description_
    """
    auth: Auth = return_auth(**kwargs)
    response = {}
    params = default_params(**kwargs)
    params = {**FOLDER[folder], **params}
    # Verify that tag doesn't already exist
    tags_get_tag = tags_get(folder=folder, tag_name=tag_name)
    if tags_get_tag:
        raise SASEObjectExists(f"Object already exists tag={tag_name}")
    data = tags_create_data(tag_name=tag_name, **kwargs)
    print(data)
    response = prisma_request(token=auth,
                              method="POST",
                              url_type='tags',
                              params=params,
                              data=json.dumps(data),
                              verify=auth.verify)
    return response


def tags_get(folder: str, tag_name: str, **kwargs) -> dict:
    """Retrieve a Tag if it exists empty dict if none is found

    Args:
        folder (str): _description_
        tag_name (str): _description_

    Returns:
        dict: _description_
    """
    # Will now loop through get all tags offseting by 500 each time
    auth: Auth = return_auth(**kwargs)
    response = {}
    tag_get_list = tags_list(folder=folder, limit=500, auth=auth)
    tag_get_list = tag_get_list['data']
    for tag in tag_get_list:
        if tag_name == tag['name']:
            response = tag
            print(f"INFO: Found Tag: {response}")
            break
    return response


def tags_create_data(tag_name: str, **kwargs) -> dict:
    """Creates tag Data Structure

    Args:
        tag_name (str): Name for Tag
        tag_color (str, Optional): Color for tag, must be part of predefined list
        tag_comments (str, Optional): Comment for Tag

    Returns:
        dict: data structure for rest call
    """
    data = {'name': tag_name}
    if kwargs.get('tag_comments'):
        data.update({'comments': kwargs['tag_comments']})
    if kwargs.get('tag_color') and kwargs.get('tag_color') in TAG_COLORS:
        data.update({'color': kwargs['tag_color']})
    return data


def tags_exist(tag_list: list, folder: str, **kwargs) -> bool:
    """Verifies if tag exists in configs

    Args:
        tag_list (list): _description_
        folder (str): _description_

    Raises:
        SASEError: _description_

    Returns:
        bool: _description_
    """
    if not isinstance(tag_list, list):
        raise SASEError(f"message=\"requires a list of tagnames\"|{tag_list=}")
    for tag in tag_list:
        if not tags_get(tag_name=tag, folder=folder, **kwargs):
            # print(f"DEBUG: {tag=} doesnot exist")
            return False
    return True


def tags_get_by_id(tag_id: str, folder: str, **kwargs) -> dict:
    """Get Tag by ID

    Args:
        tag_id (str): Requires TAG ID
        folder (str): Currently requires folder to be defined

    Returns:
        dict: _description_
    """
    auth: Auth = return_auth(**kwargs)
    params = FOLDER[folder]
    response = prisma_request(token=auth,
                              method="POST",
                              url_type='tags',
                              params=params,
                              get_object=f'/{tag_id}',
                              verify=auth.verify)
    return response



# pylint: disable=no-member
"""Configuration Management Calls"""

import datetime
import json
import time

import orjson

from prismasase import return_auth

from prismasase.configs import Auth
from prismasase.exceptions import SASEBadParam, SASECommitError
from prismasase.restapi import prisma_request
from prismasase.utilities import check_items_in_list


def config_manage_list_versions(limit: int = 50, offset: int = 0, **kwargs):
    """List the Candidate Configurations

    Args:
        limit (int, optional): _description_. Defaults to 50.
        offset (int, optional): _description_. Defaults to 0.

    Returns:
        _type_: _description_
    """
    auth: Auth = return_auth(**kwargs)
    response = prisma_request(token=auth,
                              method='GET',
                              url_type='config-versions',
                              offset=offset,
                              limit=limit,
                              verify=auth.verify)
    return response


def config_manage_rollback(**kwargs) -> dict:
    """Rollback to the running configuration; undoes all staged configs

    Returns:
        _type_: _description_
    """
    auth: Auth = return_auth(**kwargs)
    response = prisma_request(token=auth,
                              method='DELETE',
                              url_type='config-versions',
                              delete_object='/candidate',
                              verify=auth.verify)
    return response


def config_manage_push(folders: list, description: str = "No Description Provided", **kwargs):
    """Push the Candidate Configuration

    Args:
        folders (list): _description_
        description (str, optional): _description_. Defaults to "No Description Provided".

    Raises:
        SASEBadParam: _description_

    Returns:
        _type_: _description_
    """
    auth: Auth = return_auth(**kwargs)
    # verify the folders list
    show_run = config_manage_show_run(auth=auth)
    folders_valid: list = [device['device'] for device in show_run['data']]
    print(f"INFO: pushing candiate config for {str(', '.join(folders))}")
    if not check_items_in_list(list_of_items=folders, full_list=folders_valid):
        raise SASEBadParam(f"Invalid list of folders {str(', '.join(folders))}")
    # Set up json body
    data = {
        "folders": folders,
        "description": description
    }
    response = prisma_request(token=auth,
                              method='POST',
                              url_type='config-versions',
                              post_object='/candidate:push',
                              data=json.dumps(data),
                              verify=auth.verify)
    print(f"INFO: response={orjson.dumps(response).decode('utf-8')}")
    return response


def config_manage_show_run(**kwargs) -> dict:
    """Show the running configuratio

    Args:
        auth (Auth, Required): authorization required or must use configs auth

    Returns:
        dict: _description_
    """
    auth: Auth = return_auth(**kwargs)
    response = prisma_request(token=auth,
                              method='GET',
                              url_type='config-versions',
                              get_object='/running',
                              verify=auth.verify)
    return response


def config_manage_commit_subjobs(job_id: str, **kwargs) -> list:
    """Check subjobs that were created possibly after Parent Push completes

    Args:
        job_id (str): _description_

    Returns:
        list: _description_
    """
    # list out all jobs then check any job that is higher than the value that you just pushed
    # once that is done than you will need to check the status of each sub commit job
    auth: Auth = return_auth(**kwargs)
    config_jobs_list = []
    config_jobs = config_manage_list_jobs(auth=auth)
    for jobs in config_jobs['data']:
        if int(jobs['id']) > int(job_id):
            config_jobs_list.append(jobs['id'])
    return config_jobs_list


def config_manage_get_config(version_num: str, **kwargs) -> dict:
    """Get configuration by version number

    Args:
        version_num (str): _description_

    Returns:
        dict: _description_
    """
    auth: Auth = return_auth(**kwargs)
    response = prisma_request(token=auth,
                              method='GET',
                              url_type='config-versions',
                              post_object=f'/{version_num}',
                              verify=auth.verify)
    return response


def config_manage_load(version_num: str, **kwargs) -> dict:
    """Load a configuration by version number

    Args:
        version_num (str): _description_

    Returns:
        dict: _description_
    """
    auth: Auth = return_auth(**kwargs)
    data = {
        'version': version_num
    }
    response = prisma_request(token=auth,
                              method='POST',
                              url_type='config-versions',
                              data=json.dumps(data),
                              post_object=':load',
                              verify=auth.verify)
    return response


def config_manage_list_jobs(limit: int = 5, offset: int = 0, **kwargs) -> dict:
    """List configuration Jobs

    Returns:
        dict: _description_
    """
    auth: Auth = return_auth(**kwargs)
    params = {
        'limit': limit,
        'offset': offset
    }
    response = prisma_request(token=auth,
                              method='GET',
                              url_type='jobs',
                              params=params,
                              verify=auth.verify)
    return response


def config_manage_list_job_id(job_id: str, **kwargs) -> dict:
    """List configuration Job by ID Status Information

    Args:
        job_id (str): _description_

    Returns:
        dict: _description_
    """
    auth: Auth = return_auth(**kwargs)
    response = prisma_request(token=auth,
                              method='GET',
                              url_type='jobs',
                              get_object=f'/{job_id}',
                              verify=auth.verify)
    return response


def config_check_job_id(job_id: str, timeout: int = 2700, interval: int = 30, **kwargs) -> dict:
    """Used to continual check on job id status

    Args:
        job_id (str): _description_
        timeout (int, optional): Prevents infinite loop. Defaults to 2700s (45minutes).
        interval (int, optional): How often to check

    Returns:
        dict: _description_
    """
    auth: Auth = return_auth(**kwargs)
    ending_time = time.time() + timeout
    start_time = datetime.datetime.now()
    status = ""
    response = {
        'status': 'failure',
        'job_id': {str(job_id): {}},
    }
    config_job_check = config_manage_list_job_id(job_id=job_id, auth=auth)
    status = config_job_check['data'][0]['status_str']
    results = config_job_check['data'][0]['result_str']
    # TODO: When a push is done a CommitAll is set up once that
    #  is completed multiple jobs get spawned and you than have to look
    # for the new jobs and monitor tose for the commit portion of the
    # script as that is not returned in the response
    while time.time() < ending_time:
        if status == 'FIN' and results == 'OK':
            response['status'] = 'success'
            response['job_id'][str(job_id)] = config_job_check['data'][0]
            delta = datetime.datetime.now() - start_time
            response['job_id'][str(job_id)]['total_time'] = str(delta.seconds)
            print("INFO: Push returned success")
            # print(f"DEBUG: response={orjson.dumps(config_job_check['data'][0]).decode('utf-8')}")
            break
        if status == 'FIN' and results == 'FAIL':
            response['status'] = 'failure'
            response['job_id'][str(job_id)] = config_job_check['data'][0]
            delta = datetime.datetime.now() - start_time
            response['job_id'][str(job_id)]['total_time'] = str(delta.seconds)
            # print(f"DEBUG: response={orjson.dumps(config_job_check['data'][0]).decode('utf-8')}")
            break
        # print(f"DEBUG: response={orjson.dumps(config_job_check['data'][0]).decode('utf-8')}")
        time.sleep(interval)
        config_job_check = config_manage_list_job_id(job_id=job_id, auth=auth)
        status = config_job_check['data'][0]['status_str']
        results = config_job_check['data'][0]['result_str']
    return response


def config_commit(folders: list,  # pylint: disable=too-many-locals
                  description: str = "No description Provided",
                  timeout: int = 2700,
                  **kwargs) -> dict:
    """Monitor Commit Job for error or success

    Args:
        folders (list): _description_
        description (str, optional): _description_. Defaults to "No description Provided".
        timeout (int, optional): _description_. Defaults to 2700.

    Raises:
        SASECommitError: _description_
        SASECommitError: _description_

    Returns:
        _type_: _description_
    """
    auth: Auth = return_auth(**kwargs)
    # print(f"DEBUG: {kwargs=}|{auth=}")
    response = {
        'status': 'error',
        'message': '',
        'parent_job': '',
        'version_info': []
    }
    config_job_subs = []
    version = ''
    # initial push of configurations
    config_job = config_manage_push(folders=folders, description=description, auth=auth)
    if 'success' in config_job and config_job.get('success'):
        job_id = config_job['job_id']
        message = config_job['message']
        response['status'] = 'success'
        response['message'] = message
        response['parent_job'] = str(job_id)
        print(f"INFO: Pushed successfully {job_id=}|{message=}")
        # Check original push appends it to response
        response_config_check_job = config_check_job_id(job_id=job_id, timeout=timeout, auth=auth)
        response = {**response, **response_config_check_job}
        # print(f"DEBUG: Current Response {orjson.dumps(response).decode('utf-8')}")
        if response['status'] not in ['success']:
            raise SASECommitError(
                f"Intial Push failure message=\"{orjson.dumps(response).decode('utf-8')}\"")
    else:
        raise SASECommitError(
            f"Error with Push message=\"{orjson.dumps(config_job).decode('utf-8')}\"")
    # Once that commit is completed there may be additional sub jobs
    time.sleep(5)  # Provide time to create children
    config_job_subs = config_manage_commit_subjobs(job_id=job_id, auth=auth)
    print(f"INFO: Additional job search returned Jobs {','.join(config_job_subs)}")
    # TODO: fix checking subjob as not returning error properly
    if config_job_subs:
        # TODO: Multithread this as each job runs in parallel and isnt' giving the full picture
        # Once the status is not success exit out and return the error as it's a problem
        count = len(config_job_subs)
        while response['status'] == 'success' and count > 0:
            for job in config_job_subs:
                print(f"INFO: Checking on job_id {job}")
                # uses this to append each job id to the existing job to keep all info
                response_config_check_job = config_check_job_id(
                    job_id=job, timeout=timeout, auth=auth)
                response_jobs = {**response['job_id'], **response_config_check_job['job_id']}
                # Adjust depending if error is returned
                if response_config_check_job['status'] != 'success':
                    response['status'] = response_config_check_job['status']
                    response['message'] = (
                        response['message'] + f", jobid {str(job)}: " +
                        f"{response_config_check_job['job_id'][str(job)]['details']}")
                response['job_id'] = response_jobs
                # print(f"DEBUG: Current Response {orjson.dumps(response).decode('utf-8')}")
                count -= 1
    print(f"INFO: Gathering Current Commit version for tenant {auth.tsg_id}")
    # Do not send KWARGS becuase it has another auth in it possibly since auth
    # is already extracted at the top
    show_version = config_manage_show_run(auth=auth)
    # Only pull one version
    # TODO: Decide if we want to pull each version and display,
    # but since we are commiting all all versions should equal
    for obj in show_version["data"]:
        version = obj['version']
        break
    print(f"INFO: Current Running configurations are {str(version)}")
    response['version_info'] = show_version['data']
    print(f"INFO: Final Response:\n{json.dumps(response, indent=4)}")
    return response



"""Rest Calls"""

from typing import Any, Dict
import requests
import orjson

from prismasase.configs import Auth, refresh_token
from prismasase import config
from prismasase.exceptions import SASEBadRequest, SASEMissingParam


@refresh_token
def prisma_request(token: Auth, **kwargs) -> Dict[str, Any]: # pylint: disable=too-many-locals
    """_summary_

    Args:
        token (Auth): Auth class that is used to refresh bearer token upon expiration.
        url_type (str): specify the api call
        method (str): specifies the type of HTTPS method used
        params (dict, optional): specifies parameters passed to request
        data (str, optional): specifies the data being sent
        verify (str|bool, optional): sets request to verify with a custom
         cert bypass verification or verify with standard library. Defaults to True
        timeout (int, optional): sets API call timeout. Defaults to 60
        delete_object (str, required|optional): Required if method is DELETE
        put_object (str, required|optional): Required if method is PUT
        limit (int, Optional): The maximum number of results
        offset (int, Optional): The offset of the result entry
        name (string, Optional): The name of the entry
        potition (str, Optional|Required): Required if inspecting Security Rules
        get_object (str, Optional): Used if method is "GET", but additional path parameters required
    Returns:
        _type_: _description_
    """
    try:
        url_type: str = kwargs['url_type']
        method: str = kwargs['method'].upper()
    except KeyError as err:
        raise SASEMissingParam(str(err)) # pylint: disable=raise-missing-from
    params: dict = kwargs.get('params', {})
    try:
        url: str = config.REST_API[url_type]
    except KeyError as err:
        raise SASEMissingParam(f'incorrect url type: {str(err)}') # pylint: disable=raise-missing-from
    if kwargs.get('name'):
        params.update({'name': kwargs.get('name')})
    if kwargs.get('limit'):
        params.update({'limit': int(kwargs.get('limit', config.LIMIT))})
    if kwargs.get('offset'):
        params.update({'offset': int(kwargs.get('offset', config.OFFSET))})
    url: str = config.REST_API[url_type]
    headers = {"authorization": f"Bearer {token}", "content-type": "application/json"}
    data: str = kwargs.get('data', None)
    verify = kwargs.get('verify', True)
    timeout: int = kwargs.get('timeout', 90)
    if method.lower() == 'delete':
        delete_object = kwargs['delete_object']
        url = f"{url}{delete_object}"
    if method.lower() == 'put':
        put_object = kwargs['put_object']
        url = f"{url}{put_object}"
    if method.lower() == 'post' and kwargs.get('post_object'):
        post_object = kwargs['post_object']
        url = f"{url}{post_object}"
    if method.lower() == 'get' and kwargs.get('get_object'):
        get_object = kwargs['get_object']
        url = f"{url}{get_object}"
    response = requests.request(method=method,
                                url=url,
                                headers=headers,
                                data=data,
                                params=params,
                                verify=verify,
                                timeout=timeout)
    if '_errors' in response.json():
        raise SASEBadRequest(orjson.dumps(response.json()).decode('utf-8'))  # pylint: disable=no-member
    if response.status_code == 404:
        print(response.json())
        print('fail')
    if response.status_code == 400:
        return response.json()
    response.raise_for_status()
    return response.json()



"""Static Collections"""

URL_BASE: str = "https://api.sase.paloaltonetworks.com/sse/config/v1"
REMOTE_FOLDER: dict = {"folder": "Remote Networks"}
MOBILE_FOLDER: dict = {"folder": "Mobile Users"}
SERVICE_FOLDER: dict = {"folder": "Service Connections"}
SHARED_FOLDER: dict = {"folder": "Shared"}

FOLDER: dict = {
    'Remote Networks': REMOTE_FOLDER,
    'Mobile Users': MOBILE_FOLDER,
    'Service Connections': SERVICE_FOLDER,
    'Shared': SHARED_FOLDER
}

# TAG Statics
TAG_COLORS = [
    'Red', 'Green', 'Blue', 'Yellow', 'Copper', 'Orange', 'Purple', 'Gray', 'Light', 'Green', 'Cyan',
    'Light', 'Gray', 'Blue', 'Gray', 'Lime', 'Black', 'Gold', 'Brown', 'Olive', 'Maroon',
    'Red-Orange', 'Yellow-Orange', 'Forest', 'Green', 'Turquoise', 'Blue', 'Azure', 'Blue',
    'Cerulean', 'Blue', 'Midnight', 'Blue', 'Medium', 'Blue', 'Cobalt', 'Blue', 'Violet', 'Blue',
    'Blue', 'Violet', 'Medium', 'Violet', 'Medium', 'Rose', 'Lavender', 'Orchid', 'Thistle', 'Peach',
    'Salmon', 'Magenta', 'Red', 'Violet', 'Mahogany', 'Burnt', 'Sienna', 'Chestnut']

# AUTO TAG STATICS
AUTOTAG_ACTIONS = ["add-tag", "remove-tag"]
AUTOTAG_TARGET = ["source-address", "destination-address", "user", "xff-address"]
AUTOTAG_LOG_TYPE = ["traffic", "wildfire", "threat", "url", "data", "tunnel", "auth"]

# IKE Static
DYNAMIC = {'dynamic': {}}



"""Auto Tag Actions"""

import json

from prismasase import return_auth
from prismasase.configs import Auth
from prismasase.exceptions import (SASEAutoTagError, SASEAutoTagExists, SASEAutoTagTooLong)
from prismasase.utilities import (default_params, check_name_length)
from prismasase.statics import (AUTOTAG_ACTIONS, AUTOTAG_LOG_TYPE,
                                AUTOTAG_TARGET, FOLDER, SHARED_FOLDER)
from prismasase.restapi import prisma_request
from .tags import tags_get


def auto_tag_list(**kwargs) -> dict:
    """List auto Tag Actions

    Returns:
        dict: _description_
    """
    auth: Auth = return_auth(**kwargs)
    params = default_params(**kwargs)
    params = {**params, **SHARED_FOLDER}
    # Check for a named tag that may already exist
    if kwargs.get('name'):
        params = {**params, **{'name': kwargs['name']}}
    response = prisma_request(token=auth,
                              method="GET",
                              url_type='auto-tag-actions',
                              params=params,
                              verify=auth.verify)
    return response


def auto_tag_create(name: str, tag_filter: str, actions: list, **kwargs) -> dict:
    """Create an Auto Tag

    Args:
        name (str): _description_
        filter (str): _description_
        actions (list): _description_

    Raises:
        SASEAutoTagExists: _description_

    Returns:
        dict: _description_
    """
    auth: Auth = return_auth(**kwargs)
    params = SHARED_FOLDER
    # Confirm doesn't already exist
    response = auto_tag_list(name=name)
    if len(response['data']) > 0:
        # print(f"DEBUG: {response=}")
        raise SASEAutoTagExists(f"Auto Tag Already exists {name}={response['data'][0]}")
    data = auto_tag_payload(tag_filter=tag_filter, name=name, actions=actions, **kwargs)
    response = prisma_request(token=auth,
                              method='POST',
                              url_type='auto-tag-actions',
                              params=params,
                              data=json.dumps(data),
                              verify=auth.verify)
    return response


def auto_tag_payload(tag_filter: str, name: str, actions: list, log_type: str, **kwargs) -> dict:
    """Creates Tagging payload
    Sample:
    {
        "actions": [
            {
                "name": "string",
                "type": {
                    "tagging": {
                        "action": "add-tag",
                        "tags": [
                            "string"
                        ],
                        "target": "source-address",
                        "timeout": 0
                    }
                }
            }
        ],
        "description": "string",
        "filter": "string",
        "name": "string",
        "quarantine": true,
        "send_to_panorama": true,
        "folder": "Shared",
        "log_type": "traffic"
    }

    Args:
        filter (str): _description_
        name (str): _description_
        log_type (str, Requirements): log type required acceptable values

    Returns:
        dict: _description_
    """
    if not check_name_length(name=tag_filter, length=2047):
        raise SASEAutoTagTooLong(f"message=\"greater than allowed 2047\"|filter=\"{filter}\"")
    if not check_name_length(name=name, length=63):
        raise SASEAutoTagTooLong(f"message=\"greater than allowed 63\"|{name=}")
    # will raise an error if anything is missing
    auto_tag_confirm_actions(actions=actions)
    if log_type not in AUTOTAG_LOG_TYPE:
        raise SASEAutoTagError(f"message=\"log_type {log_type} not a valid type\"")
    data = {
        "filter": tag_filter,
        "name": name,
        "actions": actions,
        "log_type": log_type
    }
    if kwargs.get('folder') and isinstance(
            kwargs.get('folder'),
            str) and kwargs.get('folder') in FOLDER:
        data['folder'] = kwargs['folder']
    if kwargs.get('description'):
        data.update({'description': kwargs['description']})
    if kwargs.get('quarantine') and isinstance(kwargs.get('quarentine'), bool):
        data.update({'quarantine': kwargs['quarantine']})
    if kwargs.get('send_to_panorama') and isinstance(kwargs.get('send_to_panorama'), bool):
        data.update({'send_to_panorama': kwargs['send_to_panorama']})
    # print(f"DEBUG: {data=}")
    return data


def auto_tag_confirm_actions(actions: list):
    """Verifiy if action has all the correct parameters otherwise will raise issue.

    Args:
        actions (list): _description_

    Raises:
        SASEAutoTagError: _description_
        SASEAutoTagError: _description_
        SASEAutoTagError: _description_
        SASEAutoTagTooLong: _description_
        SASEAutoTagTooLong: _description_
        SASEAutoTagTooLong: _description_
    """
    try:
        for action in actions:
            name: str = action['name']
            tag_action: str = action['type']['tagging']['action']
            if tag_action not in AUTOTAG_ACTIONS:
                raise SASEAutoTagError(f"message=\"invalid tag action\"|{tag_action=}")
            target: str = action['type']['tagging']['target']
            tags: list = action['type']['tagging']['tags'] if action['type']['tagging'].get('tags') else [
            ]
            #timeout: int = action['type']['tagging']['timeout'] if action['type']['tagging'].get('timeout') else 0
            if not isinstance(tags, list):
                raise SASEAutoTagError(f"message=\"invalid tags type must be list\"|{tags=}")
            if target not in AUTOTAG_TARGET:
                raise SASEAutoTagError(f"message=\"invalid target\"|{target=}")
            if not check_name_length(name=name, length=63):
                raise SASEAutoTagTooLong(f"message=\"greater than allowed 63\"|{name=}")
            if len(tags) <= 64:
                raise SASEAutoTagTooLong(
                    f"message=\"list of tags too long\"|tags=\"{','.join(tags)}\"")
            for tag in tags:
                if not check_name_length(name=tag, length=127):
                    raise SASEAutoTagTooLong(f"message=\"tag name is too long\"|{tag=}")
                # check Tag exists
                tag_exists = tags_get(folder='Shared', tag_name=tag)
                # if not tag_exists['data']:
                if not tag_exists:
                    print(f"DEGUG: Tag not found {tag_exists=}")
                    raise SASEAutoTagError(f"message=\"tag doesnot exist\"|{tag=}")
    except KeyError as err:
        error = f"{type(err).__name__}: {str(err)}" if err else ""
        raise SASEAutoTagError(f"message=\"missing action value\"|{error=}") # pylint: disable=raise-missing-from
    # if all chekcks passed than it's a valid action


def auto_tag_update_filter():
    pass


def auto_tag_add_action():
    pass


def auto_tag_delete():
    pass



# pylint: disable=duplicate-key
"""IKE Utilities"""

import json
import orjson

from prismasase import return_auth
from prismasase.configs import Auth
from prismasase.exceptions import SASEBadParam, SASEBadRequest, SASEMissingParam
from prismasase.restapi import prisma_request
from prismasase.statics import DYNAMIC


def ike_gateway(pre_shared_key: str,
                ike_crypto_profile: str,
                ike_gateway_name: str,
                folder: dict,
                **kwargs) -> dict:
    """Reviews if an IKE Gateway exists or not and overwrites
     it with new configs or creates the IKE Gateway.

    Args:
        pre_shared_key (str): _description_
        ike_gateway_name (str): _description_
        local_fqdn (str): Required if peer_id_type == 'ufqdn'
        peer_fqdn (str): Required if peer_id_type == 'ufqdn'
        ike_crypto_profile (str): _description_
        peer_id_type (str): Requires one of 'ipaddr'|'fqdn'|'keyid'|'ufqdn'
        local_id_type (str): Requires one of 'ipaddr'|'fqdn'|'keyid'|'ufqdn'
    """
    auth: Auth = return_auth(**kwargs)
    params = folder
    ike_gateway_exists: bool = False
    ike_gateway_id: str = ""
    response = {}
    # data = {
    #    "authentication": {
    #        "pre_shared_key": {"key": pre_shared_key}},
    #    "local_id": {
    #        'type': 'ufqdn',
    #        'id': local_fqdn
    #    },
    #    "name": ike_gateway_name,
    #    "peer_address": {'dynamic': {}},
    #    "peer_id": {
    #        'type': 'ufqdn',
    #        'id': peer_fqdn
    #    },
    #    'protocol_common': {
    #        'nat_traversal': {'enable': True},
    #        'fragmentation': {'enable': False}
    #    },
    #    'protocol': {
    #        'ikev1': {
    #            'ike_crypto_profile': ike_crypto_profile,
    #            'dpd': {'enable': True}
    #        },
    #        'ikev2': {
    #            'ike_crypto_profile': ike_crypto_profile,
    #            'dpd': {'enable': True}
    #        },
    #        'version': 'ikev2-preferred'
    #    },
    #    "protocol_common": {
    #        "fragmentation": {"enable": False},
    #        "nat_traversal": {"enable": True},
    #        "passive_mode": True
    #    }
    # }
    data = create_ike_gateway_payload(pre_shared_key=pre_shared_key,
                                      ike_gateway_name=ike_gateway_name,
                                      ike_crypto_profile=ike_crypto_profile,
                                      **kwargs)
    # Get all current IKE Gateways
    ike_gateways = prisma_request(token=auth,
                                  url_type='ike-gateways',
                                  method='GET',
                                  params=params,
                                  verify=auth.verify)
    # Check if ike_gateway already exists
    for ike_gw in ike_gateways['data']:
        if ike_gw['name'] == ike_gateway_name:
            ike_gateway_exists = True
            ike_gateway_id = ike_gw['id']
    # Run function based off information above
    if not ike_gateway_exists:
        response = ike_gateway_create(data=data, folder=folder, **kwargs)
    else:
        response = ike_gateway_update(data=data,
                                      ike_gateway_id=ike_gateway_id,
                                      folder=folder,
                                      **kwargs)
    # print(f"DEBUG: IKE Gateway {response=}")
    return response


def ike_gateway_update(data: dict, ike_gateway_id: str, folder: dict, **kwargs) -> dict:
    """Updates an existing IKE Gateway based on the ID

    Args:
        data (dict): _description_
        ike_gateway_id (str): _description_
        auth (Auth): if provided used otherwise new one created

    Raises:
        SASEBadRequest: _description_
    """
    auth: Auth = return_auth(**kwargs)
    print(f"INFO: Updating IKE Gateway: {data['name']}")
    # print(f"DEBUG: Updating IKE Gateway Using data={json.dumps(data)}")
    params = folder
    response = prisma_request(token=auth,
                              method='PUT',
                              url_type='ike-gateways',
                              data=json.dumps(data),
                              params=params,
                              verify=auth.verify,
                              put_object=f'/{ike_gateway_id}')
    # print(f"DEBUG: response={response}")
    if '_errors' in response:
        raise SASEBadRequest(orjson.dumps(response).decode('utf-8'))  # pylint: disable=no-member
    return response


def ike_gateway_create(data: dict, folder: dict, **kwargs) -> dict:
    """_summary_

    Args:
        data (dict): ike configuration payload

    Raises:
        SASEBadRequest: _description_
    """
    auth: Auth = return_auth(**kwargs)
    print(f"INFO: Creating IKE Gateway: {data['name']}")
    # print(f"data={json.dumps(data)}")
    # print(f"DEBUG: Creating IKE Gateway Using data={json.dumps(data)}")
    params = folder
    response = prisma_request(token=auth,
                              method='POST',
                              url_type='ike-gateways',
                              data=json.dumps(data),
                              params=params,
                              verify=auth.verify)
    # print(f"DEBUG: response={response}")
    if '_errors' in response:
        raise SASEBadRequest(orjson.dumps(response).decode('utf-8'))  # pylint: disable=no-member
    return response


def create_ike_gateway_payload(pre_shared_key: str,
                               ike_crypto_profile: str,
                               ike_gateway_name,
                               **kwargs) -> dict:
    """Creates IKE Gateway Payload used to create an IKE Gateway.
    Uses format "ike-gw-<remote_network_name>"

    Args:
        pre_shared_key (str): _description_
        local_fqdn (str): _description_
        peer_fqdn (str): _description_
        ike_crypto_profile (str): _description_

    Returns:
        dict: data payload
    """
    try:
        peer_address_type = kwargs.get('peer_address_type', 'dynamic')
        if peer_address_type.lower() == 'dynamic':
            # Sets peer_address to Dynamic Object if set to 'dynamic'
            peer_address = DYNAMIC
        else:
            peer_address = kwargs['peer_address']
            if peer_address == 'dynamic':
                raise SASEBadParam(
                    f"message=\"peer address does not match type\"|{peer_address=}|{peer_address_type=}")
        local_id_type: str = kwargs.get('local_id_type', 'ufqdn')
        peer_id_type: str = kwargs.get('peer_id_type', 'ufqdn')
        # TODO: Fix this so that the value being passed is dynamic based on the id_type
        peer_id_value: str = kwargs['peer_id_value']
        local_id_value: str = kwargs['local_id_value']
        passive_mode: bool = bool(kwargs.get('passive_mode', 'true').lower() in ['true'])
        fragmentation: bool = bool(kwargs.get('fragmentation', 'false').lower() in ['true'])
        nat_traversal: bool = bool(kwargs.get('fragmentation', 'true').lower() in ['true'])
        # TODO: Fix protocol version currently only really working with ikev2-preffered so hardcoded
        ike_protocol_version: str = kwargs.get('ike_protocol_version', "ikev2-preferred")
    except KeyError as err:
        raise SASEMissingParam(f"message=\"missing required parameter\"|param={str(err)}")
    data = {
        "authentication": {
            "pre_shared_key": {"key": pre_shared_key}},
        "local_id": {
            'type': local_id_type,
            'id': local_id_value
        },
        "name": ike_gateway_name,
        "peer_address": peer_address,
        "peer_id": {
            'type': peer_id_type,
            'id': peer_id_value
        },
        'protocol_common': {
            'nat_traversal': {'enable': nat_traversal},
            'fragmentation': {'enable': fragmentation}
        },
        'protocol': {
            'ikev1': {
                'ike_crypto_profile': ike_crypto_profile,
                'dpd': {'enable': True}
            },
            'ikev2': {
                'ike_crypto_profile': ike_crypto_profile,
                'dpd': {'enable': True}
            },
            'version': ike_protocol_version
        },
        "protocol_common": {
            "fragmentation": {"enable": fragmentation},
            "nat_traversal": {"enable": nat_traversal},
            "passive_mode": passive_mode
        }
    }
    # data = {
    #    "authentication": {
    #        "pre_shared_key": {"key": pre_shared_key}
    #    },
    #    "local_id": {
    #        "type": "ufqdn",
    #        "id": local_fqdn
    #    },
    #    "name": ike_gateway_name,
    #    "peer_address": {"dynamic": {}},
    #    "peer_id": {
    #        "type": "ufqdn",
    #        "id": peer_fqdn
    #    },
    #    "protocol": {
    #        "ikev2": {
    #            "ike-crypto-profile": ike_crypto_profile,
    #            "dpd": {"enable": True}
    #        },
    #        "ikev1": {
    #            "ike-crypto-profile": ike_crypto_profile,
    #            "dpd": {"enable": True}
    #        },
    #        "version": "ikev2-preferred"
    #    },
    #    "protocol_common": {
    #        "fragmentation": {"enable": False},
    #        "nat_traversal": {"enable": True},
    #        "passive_mode": True
    #    }
    # }
    return data



# pylint: disable=raise-missing-from
"""Remote Networks"""

from typing import Any, Dict, List
import ipaddress
import json
import orjson

from prismasase import return_auth

from prismasase.configs import Auth
from prismasase.exceptions import (
    SASEBadParam, SASEBadRequest, SASEMissingIkeOrIpsecProfile, SASEMissingParam,
    SASENoBandwidthAllocation)
from prismasase.restapi import prisma_request
from prismasase.statics import FOLDER, REMOTE_FOLDER
from prismasase.utilities import set_bool
from ..ipsec.ipsec_tun import ipsec_tunnel
from ..ipsec.ipsec_crypto import ipsec_crypto_profiles_get
from ..ike.ike_crypto import ike_crypto_profiles_get
from ..ike.ike_gtwy import ike_gateway


def bulk_import_remote_networks(remote_sites: list):
    """_summary_

    Args:
        remote_sites (list): _description_
    """
    # TODO: build in a way to bulk import if list
    # or dictionary is fead into this process
    pass


def create_remote_network(**kwargs) -> Dict[str, Any]:  # pylint: disable=too-many-locals
    """Creates a Remote Nework IPSec Tunnel

    Args:
        remote_network_name (str): Remote Network Name used to set up Remote Network
        region (str): region checking for bandwidth allocation
        spn_name (str): IPSec Termination Node name
        pre_shared_key (str, Required): required to set a pre-shared-key
        ike_crypto_profile (str): IKE Crypto Profile to use (must exist in tenant)
        ipsec_crypto_profile (str): IPSec Crypto profile to use (must exist in tenant)
        ike_gateway_name (str, Optional): The IKE Gateway Name. Default: ike-gw-{remote_network_name}
        ipsec_tunnel_name (str, Optional): The IPsec Tunnel Name. Default: ipsec-tunnel-{remote_network_name}
        auth (Auth, Optional): Authorization if none supplied it defaults to the Yaml Config
        fragmentation (str): IKE Protocol Fragmentation. Default False
        nat_traversal (str): IKE Protocal Allow NAT Traversal. Default True
        peer_id_type (str): Requires one of 'ipaddr'|'fqdn'|'keyid'|'ufqdn'. Defaults 'ufqdn
        peer_id_value (str): Dependign on peer_id_type this requires the corresponding value
        peer_address_type (str): Requires one of 'ipaddr'|'fqdn'|'dynamic'
        peer_address (str): Requires literal "dynamic" for ufqdn or ip_address or fqdn_name or key if other is selected as peer_id_type
        local_id_type (str): Requires one of 'ipaddr'|'fqdn'|'keyid'|'ufqdn'. Defaults 'ufqdn'
        local_id_value (str): Depending on the local_id_type will determine the type of value needed
        passive_mode (str): Sets IKE passive mode deafaults to 'true'
        bgp_enabled (str|bool): Sets BGP routing enabled or disabled use string 'true' or 'false' Defaults 'false'
        bgp_peer_ip (str): Required if bgp_enabled is 'true'
        bgp_local_ip (str): Required if bgp_enabled is 'true'
        bgp_peer_as (str): Required if bgp_enabled is 'true'
        static_enabled (str|bool): Sets Static routing enabled or disabled use string 'true' or 'false' Defaults 'false'
        tunnel_monitor (str|bool): Sets Tunnel Monitoring to enabled or disabled use string 'true' or 'false' Defaults 'false'


    Raises:
        SASEMissingParam: _description_
        SASENoBandwidthAllocation: _description_
    """
    response = {
        "status": "error",
        "message": {},
    }
    try:
        auth: Auth = return_auth(**kwargs)
        remote_network_name: str = kwargs.pop('remote_network_name')
        region: str = kwargs.pop('region')
        spn_name: str = kwargs.pop('spn_name')
        ike_crypto_profile: str = kwargs.pop('ike_crypto_profile')
        ipsec_crypto_profile: str = kwargs.pop('ipsec_crypto_profile')
        ike_gateway_name: str = kwargs.pop('ike_gateway_name') if kwargs.get(
            'ike_gateway_name') else f"ike-gwy-{remote_network_name}"
        ipsec_tunnel_name: str = kwargs['ipsec_tunnel_name'] if kwargs.get(
            'ipsec_tunnal_name') else f"ipsec-tunnel-{remote_network_name}"
        # Converts string values to bool and passes default values
        tunnel_monitor: bool = set_bool(value=kwargs.pop('tunnel_monitor', ''), default=False)
        # monitor_ip: str = kwargs['monitor_ip'] if tunnel_monitor else ""
        static_enabled: bool = set_bool(value=kwargs.pop('static_enabled', ''), default=False)
        #static_routing: list = kwargs.pop('static_routing').split(',') if static_enabled else []
        bgp_enabled: bool = set_bool(value=kwargs.pop('bgp_enabled', ''), default=False)
        # Default folder to "Remote Networks" for reverse compatability
        pre_shared_key = kwargs.pop('pre_shared_key')
        if kwargs.get("folder_name"):
            folder: dict = FOLDER[kwargs.pop('folder_name')]
        else:
            folder: dict = kwargs['folder'] if kwargs.get('folder') else REMOTE_FOLDER
    except KeyError as err:
        raise SASEMissingParam(f"message=\"missing required parameter\"|param={str(err)}")
    # Check Bandwdith allocations
    # print(f"{region=},{spn_name=}")
    bandwidth_check = verify_bandwidth_allocations(
        name=region, spn_name=spn_name, folder=folder, auth=auth)
    if not bandwidth_check:
        raise SASENoBandwidthAllocation(
            "No Bandwidth Association or allocations exists for " + f"{region=} {spn_name=}")
    # Verify IKE and IPSec Profiles exist
    if not verify_ike_ipsec_profiles_exist(ike_crypto_profile=ike_crypto_profile,
                                           ipsec_crypto_profile=ipsec_crypto_profile,
                                           folder=folder, auth=auth):
        raise SASEMissingIkeOrIpsecProfile(
            'message=\"Missing a profile in configurations\"|' +
            f'{ike_crypto_profile=}|{ipsec_crypto_profile=}')
    print(f"INFO: Verified {region=} and {spn_name=} exist")
    # Create IKE Gateway
    print(f"INFO: IKE Gateway Name = {ike_gateway_name}")
    # Create IKE Gateway
    response_ike_gateway = ike_gateway(pre_shared_key=pre_shared_key,
                                       ike_crypto_profile=ike_crypto_profile,
                                       ike_gateway_name=ike_gateway_name,
                                       folder=folder,
                                       **kwargs)
    # print(f"DEBUG: IKE Gateway {response_ike_gateway=}")
    response['message'].update({'ike_gateway': response_ike_gateway})
    # Create IPSec Tunnel
    response_ipsec_tunnel = ipsec_tunnel(ipsec_tunnel_name=ipsec_tunnel_name,
                                         ipsec_crypto_profile=ipsec_crypto_profile,
                                         ike_gateway_name=ike_gateway_name,
                                         tunnel_monitor=tunnel_monitor,
                                         folder=folder,
                                         **kwargs)
    response['message'].update({'ipsec_tunnel': response_ipsec_tunnel})
    # print(f"DEBUG: IPSec Tunnel {response_ipsec_tunnel=}")
    # Create Remote Network
    response_remote_network = remote_network(remote_network_name=remote_network_name,
                                             ipsec_tunnel_name=ipsec_tunnel_name,
                                             region=region,
                                             spn_name=spn_name,
                                             static_enabled=static_enabled,
                                             bgp_enabled=bgp_enabled,
                                             folder=folder,
                                             **kwargs)
    response['message'].update({'remote_network': response_remote_network})
    response['status'] = 'success'
    # print(f"DEBUG: Remote Network {response_remote_network=}")
    print(f"INFO: Created Remote Network \n{json.dumps(response, indent=4)}")
    return response


def verify_bandwidth_allocations(name: str, spn_name: str, folder: dict, **kwargs) -> bool:
    """Verifies that the region has allocated bandwidth and that the spn exists

    Args:
        name (str): region checking for bandwidth allocation
        spn_name (str): IPSec Termination Node name

    Returns:
        bool: True if exists
    """
    auth: Auth = return_auth(**kwargs)
    bandwidth_check = False
    bandwidth = get_bandwidth_allocations(folder=folder, auth=auth)
    if bandwidth:
        for entry in bandwidth:
            if entry['name'].lower() in name.lower():
                if spn_name in entry['spn_name_list']:
                    bandwidth_check = True
    return bandwidth_check


def get_bandwidth_allocations(folder: dict, **kwargs) -> List[Dict[str, Any]]:
    """Gets Bandwith Allocations for Tenant

    Args:
        name (str): Region Name

    Returns:
        List[Dict[str,Any]]: List of region information and the amount of bdwth allocated
    """
    auth: Auth = return_auth(**kwargs)
    params = folder
    # print(f"DEBUG: {auth.verify}")
    bandwidth = prisma_request(token=auth,
                               url_type='bandwidth-allocations',
                               method='GET',
                               params=params,
                               verify=auth.verify)
    return bandwidth['data']


def verify_ike_ipsec_profiles_exist(ike_crypto_profile: str,
                                    ipsec_crypto_profile: str,
                                    folder: dict,
                                    **kwargs) -> bool:
    """Verifies that both IKE Profile and IPsec Profiles exist

    Args:
        ike_crypto_profile (str): _description_
        ipsec_crypto_profile (str): _description_
        auth (Auth, optional): if not provided default auth will be passed

    Returns:
        bool: _description_
    """
    return all([ike_crypto_profiles_get(
        ike_crypto_profile=ike_crypto_profile, folder=folder, **kwargs),
        ipsec_crypto_profiles_get(
        ipsec_crypto_profile=ipsec_crypto_profile, folder=folder, **kwargs)])


def remote_network(remote_network_name: str,
                   ipsec_tunnel_name: str,
                   region: str,
                   spn_name: str,
                   static_enabled: bool,
                   bgp_enabled: bool,
                   folder: dict,
                   **kwargs) -> dict:
    """Create a Remote Network

    Args:
        remote_network_name (str): _description_
        region (str): _description_
        spn_name (str): _description_
        static_enabled (bool): _description_
        bgp_enabled (bool): _description_

    Raises:
        SASEMissingParam: _description_
        SASEMissingParam: _description_
    """
    auth = return_auth(**kwargs)
    params = folder
    remote_network_exists: bool = False
    remote_network_id: str = ""
    data = create_remote_network_payload(remote_network_name=remote_network_name,
                                         ipsec_tunnel_name=ipsec_tunnel_name,
                                         region=region,
                                         spn_name=spn_name)
    if static_enabled:
        try:
            data['subnets'] = kwargs['static_routing'].split(',') if static_enabled else []
            #data["subnets"] = kwargs['static_routing']
            for subnet in data['subnets']:
                try:
                    ipaddress.ip_network(subnet)
                except ValueError:
                    raise SASEBadParam(f"message=\"incorrect IP Network\"|{subnet=}")
            if len(data['subnets']) == 0:
                raise SASEMissingParam(
                    "message=\"required subnet if static routing enabled\"" +
                    f"|param={data['subnets']}")
        except KeyError as err:
            raise SASEMissingParam(
                f'message=\"required when static_enabled is True\"|param={str(err)}')
    if bgp_enabled:
        data = create_remote_network_bgp_payload(data=data, **kwargs)
    # Check if remote network already exists
    remote_networks = prisma_request(token=auth,
                                     url_type='remote-networks',
                                     method='GET',
                                     params=params,
                                     verify=auth.verify)
    # Check if remote network already exists
    for network in remote_networks['data']:
        if network['name'] == remote_network_name:
            remote_network_exists = True
            remote_network_id = network['id']
    # Run create or update functions
    if not remote_network_exists:
        response = remote_network_create(data=data,
                                         folder=folder,
                                         auth=auth)
    else:
        response = remote_network_update(data=data,
                                         remote_network_id=remote_network_id,
                                         folder=folder,
                                         auth=auth)
    return response


def remote_network_create(data: dict, folder: dict, **kwargs) -> dict:
    """Create a new remote nework connection

    Args:
        data (dict): _description_

    Raises:
        SASEBadRequest: _description_
    """
    auth: Auth = return_auth(**kwargs)
    params = folder
    # print(f"DEBUG: remote_network_create={json.dumps(data)}")
    response = prisma_request(token=auth,
                              method='POST',
                              url_type='remote-networks',
                              data=json.dumps(data),
                              params=params,
                              verify=auth.verify)
    # print(f"DEBUG: response={response}")
    if '_errors' in response:
        raise SASEBadRequest(orjson.dumps(response).decode('utf-8'))  # pylint: disable=no-member
    return response


def remote_network_update(data: dict, remote_network_id: str, folder: dict, **kwargs) -> dict:
    """Update an existing remote network

    Args:
        data (dict): _description_
        remote_network_id (str): _description_

    Raises:
        SASEBadRequest: _description_
    """
    auth: Auth = return_auth(**kwargs)
    params = folder
    # print(f"DEBUG: remote_network_create={json.dumps(data)}")
    response = prisma_request(token=auth,
                              method='PUT',
                              data=json.dumps(data),
                              params=params,
                              url_type='remote-networks',
                              verify=auth.verify,
                              put_object=f'/{remote_network_id}')
    # print(f"DEBUG: response={response}")
    if '_errors' in response:
        raise SASEBadRequest(orjson.dumps(response).decode('utf-8'))  # pylint: disable=no-member
    return response


def remote_network_delete(remote_network_id: str, folder: dict, **kwargs) -> dict:
    """DELETE a remote network

    Args:
        remote_network_id (str): _description_

    Returns:
        dict: _description_
    """
    auth: Auth = return_auth(**kwargs)
    params = folder
    response = prisma_request(token=auth,
                              method='DELETE',
                              params=params,
                              delete_object=f'/{remote_network_id}',
                              verify=auth.verify)
    return response


def create_remote_network_bgp_payload(data: dict,
                                      **kwargs) -> dict:
    """Create BGP payload if BGP is enabled

    Args:
        data (dict): _description_
        bgp_local_ip (str): _description_
        bgp_peer_as (str): _description_
        bgp_peer_ip (str): _description_
        bgp_summarized_mobile_user_routes (str): Defaults to True
        bgp_peering_type (str): Defaults to "exchange-v4-over-v4"

    Returns:
        dict: _description_
    """
    try:
        bgp_local_ip = kwargs['bgp_local_ip']
        bgp_peer_as = kwargs['bgp_peer_as']
        bgp_peer_ip = kwargs['bgp_peer_ip']
        bgp_peering_type = kwargs.get('bgp_peering_type', "exchange-v4-over-v4")
        bgp_summarized_mobile_user_routes: bool = bool(kwargs.get(
            'bgp_summarized_mobile_user_routes', 'true').lower() in ['true'])
    except KeyError as err:
        raise SASEMissingParam(f"message=\"missing required parameter\"|param={str(err)}")
    data["protocol"] = {
        "bgp": {
            "do_not_export_routes": False,
            "enable": True,
            "local_ip_address": bgp_local_ip,
            "originate_default_route": True,
            "peer_as": bgp_peer_as, "peer_ip_address": bgp_peer_ip,
            "peering_type": bgp_peering_type,
            "summarize_mobile_user_routes": bgp_summarized_mobile_user_routes
        },
        "bgp_peer": {
            "local_ip_address": bgp_local_ip,
            "peer_ip_address": bgp_peer_ip
        }
    }
    return data


def create_remote_network_payload(**kwargs) -> dict:
    """Creates Remote Network Payload

    Args:
        remote_network_name (str): _description_
        region (str): _description_
        spn_name (str): _description_
        name (str): remote network name
        ipsec_tunnel (str): IPSec Tunnel Name
        license_type (str): License type

    Returns:
        dict: data dictionary used for body of request
    """
    try:
        data = {
            "ipsec_tunnel": kwargs['ipsec_tunnel_name'],
            "license_type": kwargs['license_type']
            if kwargs.get('license_type') else "FWAAS-AGGREGATE",
            "name": kwargs['remote_network_name'],
            "region": kwargs['region'],
            "spn_name": kwargs['spn_name']}
    except KeyError as err:
        raise SASEMissingParam(f"message=\"missing param for payload\"|param={str(err)}")
    return data


def remote_network_list(folder: dict, limit: int = 200, offset: int = 0, **kwargs) -> dict:
    """Retrieves a list of all Remote Networks

    Args:
        limit (int, optional): _description_. Defaults to 200.
        offset (int, optional): _description_. Defaults to 0.

    Returns:
        dict: _description_
    """
    auth: Auth = return_auth(**kwargs)
    params = {
        "limit": limit,
        "offset": offset
    }
    params = {**folder, **params}
    response = prisma_request(token=auth,
                              method='GET',
                              url_type='remote-networks',
                              params=params,
                              verify=auth.verify)
    return response


def remote_network_identifier(name: str, folder: dict, **kwargs) -> dict:
    """Returns Remote Newtork Data Information

    Args:
        name (str): _description_

    Returns:
        dict: _description_
    """
    auth: Auth = return_auth(**kwargs)
    response: dict = {}
    remote_net_list = remote_network_list(folder=folder, auth=auth)
    for remote_net in remote_net_list['data']:
        if name in remote_net_list['name']:
            response = remote_net
            break
    return response
